"""Image processing functions for BactoVision."""

import warnings
from typing import Dict, Tuple

import cv2 as cv
import numpy as np
from PIL import Image, ImageDraw
from scipy.spatial import ConvexHull
from skimage.filters import gaussian, threshold_otsu
from skimage.measure import label
from skimage.morphology import remove_small_objects
from skimage.segmentation import clear_border

__all__ = [
    "segment_by_thresholding",
    "get_summary_metrics",
    "clahe",
    "add_convex_hulls",
    "preprocess_image",
    "normalize_image",
    "working_preprocessing",
]


def segment_by_thresholding(
    image,
    t: float = 1.0,
    s: float = 1,
    convexhull: bool = True,
    small_obj_percent: float = 0.00005,
) -> Tuple[np.ndarray, np.ndarray]:
    """Identifies suitable morphological components from image by thresholding.

    Args:
        image: numpy array of the image.
        t: threshold for thresholding.
        s: size of the morphological component.
        convexhull: whether to add convex hulls to the mask.
        small_obj_percent: percentage of the image area to filter small components.

    Returns:
        Tuple of the original image and the mask.
    """
    # Apply Otsu thresholding
    thresh = t * threshold_otsu(image)

    mask = image > thresh

    # Filter small components. The default threshold is 0.005 % of the image area
    size_thresh = s * np.prod(image.shape) * small_obj_percent

    mask = remove_small_objects(mask, min_size=size_thresh)

    # Clear border
    mask = clear_border(mask)

    # Label connected components
    mask = label(mask)

    if convexhull:
        mask = add_convex_hulls(mask)

    return image, mask


def add_convex_hulls(mask: np.ndarray) -> np.ndarray:
    """Create convex hulls around each connected component in mask.

    Args:
        mask: numpy array of the mask.

    Returns:
        Processed mask.
    """
    labels = np.unique(mask)[1:]

    if not labels.size:
        return mask

    new_mask = np.zeros_like(mask)

    for mask_label in labels:
        region = np.argwhere(mask == mask_label)
        hull = ConvexHull(region)
        verts = [(region[v, 0], region[v, 1]) for v in hull.vertices]
        img = Image.new("L", mask.shape, 0)
        ImageDraw.Draw(img).polygon(verts, outline=1, fill=1)
        new_mask += np.array(img).T * mask_label

    return new_mask


def working_preprocessing(
    img,
    subtract_background: bool = True,
    use_clahe: bool = True,
    clahe_limit: float = 200,
) -> np.ndarray:
    """Preprocess image for thresholding and analysis.

    Args:
        img: numpy array of the image.
        subtract_background: whether to subtract the background.
        use_clahe: whether to use CLAHE.
        clahe_limit: limit for CLAHE.

    Returns:
        Preprocessed image.
    """
    img = normalize_image(np.array(img)) * 255

    if use_clahe:
        img = cv.cvtColor(img.astype(np.uint16), cv.COLOR_BGR2GRAY)
        img = (
            normalize_image(
                cv.createCLAHE(clipLimit=clahe_limit, tileGridSize=(1, 1)).apply(
                    img.astype("uint16")
                )
            )
            * 255
        )
        img = cv.applyColorMap(img.astype(np.uint8), cv.COLORMAP_CIVIDIS)

    img = img.astype(np.float32)

    img = 0.5 * img[:, :, 1] + 1 * img[:, :, 2]

    if subtract_background:
        # Estimate background by gaussian.
        # Scale sigma with image area to compensate for different resolutions
        background = gaussian(img, sigma=np.prod(img.shape) / 10000, truncate=4)
        img -= background  # This may contain some negative values

    # Scale image to [0,1] in invert
    img = 1 - normalize_image(img)

    return img


def preprocess_image(
    img: np.ndarray,
    subtract_background: bool = False,
    use_clahe: bool = False,
    clahe_limit: float = 200,
    channel_weights: tuple = (0, 0.5, 1),
    clahe_first: bool = True,
    **kwargs,
) -> np.ndarray:
    """Prepare image for thresholding and analysis.

    Channels are weighted by (0, 0.5, 1) and summed.
    The background is estimated by gaussian blur and subtracted.
    The image is scaled to [0, 1] and inverted.

    Args:
        img: numpy array of the image.
        subtract_background: whether to subtract the background.
        use_clahe: whether to use CLAHE.
        clahe_limit: limit for CLAHE.
        channel_weights: weights for the channels.
        clahe_first: whether to apply CLAHE first.
        **kwargs: additional arguments for CLAHE.

    Returns:
        Preprocessed image.
    """
    if clahe_first and use_clahe:
        img = normalize_image(img) * 255
        img = cv.cvtColor(img.astype(np.uint16), cv.COLOR_BGR2GRAY)
        img = clahe(img, limit=clahe_limit, **kwargs)
        img = cv.applyColorMap(img.astype(np.uint8), cv.COLORMAP_CIVIDIS).astype(np.float32)

    # Sum weighted channels
    img = np.sum([img[:, :, i] * w for i, w in enumerate(channel_weights)], axis=0)

    # Convert to float and rescale to range [0,1]

    if subtract_background:
        # Estimate background by gaussian.
        # Scale sigma with image area to compensate for different resolutions
        background = gaussian(img, sigma=np.prod(img.shape) / 10000, truncate=4)
        img -= background  # This may contain some negative values

    if use_clahe and not clahe_first:
        img = clahe(normalize_image(img) * 255, limit=clahe_limit, **kwargs).astype(np.float32)

    # Scale image to [0,1] and invert
    img = 1 - normalize_image(img)

    return img


def normalize_image(image: np.ndarray) -> np.ndarray:
    """Normalize image to [0, 1].

    Args:
        image: numpy array of the image.

    Returns:
        Normalized image.
    """
    image = image.astype(np.float32)
    delta = np.max(image) - np.min(image)

    if delta:
        image = (image - np.min(image)) / delta
    else:
        image = np.zeros_like(image)

    return image


def get_summary_metrics(
    image,
    mask,
    grid_x,
    grid_y,
    mode: str = "luminance-inverse",
) -> Dict[str, np.ndarray]:
    """Construct summary metrics.

    Metrics include label, area, mean intensity,
    integral opacity and average opacity for each tile.

    Args:
        image: numpy array of the image.
        mask: numpy array of the mask.
        grid_x: number of tiles in the x direction.
        grid_y: number of tiles in the y direction.
        mode: mode to calculate the brightness, one of
            'luminance', 'luminance-inverse', 'red', 'green', 'blue'.
            Default is 'luminance-inverse' used in the paper.

    Returns:
        A dictionary containing the following keys:
        - 'intergal_opacity': numpy array of the integral opacity.
        - 'average_opacity': numpy array of the average opacity.
        - 'relative_area': numpy array of the relative area.
    """
    brightness = img2brightness(image, mode=mode)

    brightness_patches = get_patches(brightness, grid_y, grid_x)
    mask_patches = get_patches(mask, grid_y, grid_x)

    num_pixels = mask_patches.sum(axis=(-1, -2))
    tile_pixes = np.prod(mask_patches.shape[-2:])

    relative_area = num_pixels / tile_pixes

    num_background_pixels = tile_pixes - num_pixels

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        c1 = 1 / num_pixels
        c1[num_pixels == 0] = 1
        c2 = 1 / num_background_pixels
        c2[num_background_pixels == 0] = 1

    average_background_brightness = (brightness_patches * (1 - mask_patches)).sum(
        axis=(-2, -1)
    ) * c2
    intergal_opacity = (brightness_patches * mask_patches).sum(
        axis=(-2, -1)
    ) - average_background_brightness * num_pixels
    average_opacity = intergal_opacity * c1

    return dict(
        intergal_opacity=intergal_opacity,
        average_opacity=average_opacity,
        relative_area=relative_area,
        num_pixels=num_pixels,
    )


def img2brightness(img: np.ndarray, mode: str = "luminance") -> np.ndarray:
    if mode == "luminance":
        return img[..., 0] * 0.2989 + img[..., 1] * 0.5870 + img[..., 2] * 0.1140
    elif mode == "sum":
        return img.sum(-1)
    elif mode == "luminance-inverse":
        img = 255 - img
        return img[..., 0] * 0.2989 + img[..., 1] * 0.5870 + img[..., 2] * 0.1140
    else:
        raise ValueError(f"Unknown mode {mode}")


def get_patches(img: np.ndarray, y_grid: int, x_grid: int) -> np.ndarray:
    if img.ndim == 2:
        reduce_channels = True
        img = img[..., None]
    else:
        reduce_channels = False

    ysize, xsize, channels = img.shape
    y_delta = ysize // y_grid
    x_delta = xsize // x_grid

    patches = img[: y_delta * y_grid, : x_delta * x_grid].reshape(
        y_grid, y_delta, x_grid, x_delta, channels
    )
    patches = patches.transpose(0, 2, 1, 3, 4)
    if reduce_channels:
        patches = patches[..., 0]
    return patches


def clahe(img: np.ndarray, limit: float = 200, grid_size: tuple = (1, 1)) -> np.ndarray:
    """Apply Contrast Limited Adaptive Histogram Equalization (CLAHE) to img.

    Args:
        img: numpy array of the image.
        limit: limit for CLAHE.
        grid_size: size of the grid for CLAHE.

    Returns:
        CLAHE-processed image.
    """
    return cv.createCLAHE(clipLimit=limit, tileGridSize=grid_size).apply(img.astype("uint16"))



"""BactoVision widget module for interactive visualization and analysis."""

from pathlib import Path
from typing import Literal, Optional, Union

import ipywidgets
import numpy as np
from ipywidgets import (
    Button,
    Dropdown,
    FloatLogSlider,
    FloatSlider,
    HBox,
    IntSlider,
    Layout,
    ToggleButton,
    ToggleButtons,
    VBox,
)
from PIL import Image

from bactovision.canvas_widget import CanvasWidget
from bactovision.image_processing import (
    get_summary_metrics,
    preprocess_image,
    segment_by_thresholding,
    working_preprocessing,
)


class BactoWidget(VBox):
    """BactoWidget is a jupyter widget for bacterial colony growth image analysis.

    Methods:
        get_metrics: get the metrics of the image after the annotation.
        get_annotation_mask: get the annotation mask.
    """

    def __init__(self, img: Union[np.ndarray, str, Path], mask: Optional[np.ndarray] = None):
        """Initialize the BactoWidget.

        Args:
            img: image to be analyzed, numpy array or path to image file.
            mask: numpy array of `mask` - an annotation with the same shape as the image,
                where the pixels with value 1 correspond to the bacterial colony, and 0
                otherwise. Optional, default is None.
        """
        # load image
        if isinstance(img, (str, Path)):
            img = np.array(Image.open(img))

        assert isinstance(img, np.ndarray), "img should be a numpy array or a path to an image"
        assert img.ndim == 3, "img should be a 3D array"

        # save original image
        self.original_img = np.array(img)

        self._saved_mask = mask
        self._prev_draw_btn_state = "Off"

        # init all widgets
        self.draw_mode_btns = ToggleButtons(
            options=["Off", "Add", "Erase"], disable=False, description="Manual annotation:"
        )

        layout_dict = dict(
            display="flex",
            flex_flow="column",
            align_items="stretch",
            style={"description_width": "initial"},
            layout=Layout(width="auto"),
        )

        self.brush_size_slider = FloatSlider(
            5, min=1, max=20, description="brush size", **layout_dict
        )
        self.hide_annotation_btn = ToggleButton(
            value=False, description="Hide annotation", **layout_dict
        )
        self.canvas_widget = CanvasWidget()
        self.threshold_slider = FloatSlider(
            1.0,
            min=0.01,
            max=2,
            step=0.001,
            description="Brightness threshold",
            continuous_update=False,
            **layout_dict,
        )
        self.small_object_size = FloatSlider(
            1.0,
            min=0.1,
            max=3.0,
            step=0.01,
            description="Smallest size",
            continuous_update=False,
            **layout_dict,
        )
        self.annotate_btn = Button(value=False, description="Apply auto", **layout_dict)
        self.clahe_btn = ToggleButton(value=False, description="Apply CLAHE", **layout_dict)
        self.clahe_limit_slider = FloatLogSlider(
            200, min=1, max=4, description="CLAHE limit", continuous_update=False, **layout_dict
        )

        self.subtract_background_btn = ToggleButton(
            value=False, description="Subtract background", **layout_dict
        )

        self.cmap_down = Dropdown(
            options=["YlGnBu", "gray", "viridis", "cividis", "plasma"],
            value="YlGnBu",
            description="Colormap",
        )

        # grid
        self.hide_grid_btn = ToggleButton(value=False, description="Hide grid", **layout_dict)
        self.change_grid_btn = ToggleButton(value=False, description="Adjust grid", **layout_dict)

        self.x_grid_size_slider = IntSlider(
            self.canvas_widget.grid_num_x, 1, 15, description="Horizontal grid size", **layout_dict
        )
        self.y_grid_size_slider = IntSlider(
            self.canvas_widget.grid_num_y, 1, 15, description="Vertical grid size", **layout_dict
        )
        self.top_pad_slider = IntSlider(self.canvas_widget.pad_top, 0, 200, **layout_dict)
        self.bottom_pad_slider = IntSlider(self.canvas_widget.pad_bottom, 0, 200, **layout_dict)
        self.right_pad_slider = IntSlider(self.canvas_widget.pad_right, 0, 200, **layout_dict)
        self.left_pad_slider = IntSlider(self.canvas_widget.pad_left, 0, 200, **layout_dict)

        self.control_widgets = [
            self.draw_mode_btns,
            self.brush_size_slider,
            self.threshold_slider,
            self.small_object_size,
            self.annotate_btn,
        ]

        # set data
        self._preprocessed_img = preprocess_image(self.cut_img(self.original_img))

        if mask is None:
            mask = np.zeros_like(self._preprocessed_img)

        self._set_image(self._preprocessed_img)
        self.canvas_widget.set_annotation(mask)

        # init connections
        self.init_connections()

        # layout

        frame_layout = Layout(
            border="1px solid gray",  # Add a solid border to create the frame
            padding="2px",  # Add some padding inside the frame
            margin="2px",  # Add some margin outside the frame
            width="auto",  # Adjust width
            height="auto",  # Adjust height based on content
        )

        fixed_frame_layout = Layout(
            border="1px solid gray",  # Add a solid border to create the frame
            padding="2px",  # Add some padding inside the frame
            margin="2px",  # Add some margin outside the frame
            width="50%",  # Adjust width
            height="auto",  # Adjust height based on content
        )

        self.grid_widgets = VBox(
            [
                HBox(
                    [
                        self.change_grid_btn,
                        self.hide_grid_btn,
                    ]
                ),
                self.x_grid_size_slider,
                self.y_grid_size_slider,
            ],
            layout=frame_layout,
        )

        self.contrast_widgets = VBox(
            [
                self.clahe_btn,
                self.clahe_limit_slider,
                self.subtract_background_btn,
                self.cmap_down,
            ],
            layout=frame_layout,
        )

        self.manual_annotation_widgets = VBox(
            [
                self.draw_mode_btns,
                self.brush_size_slider,
            ],
            layout=frame_layout,
        )

        self.auto_annotation_widgets = VBox(
            [
                HBox([self.annotate_btn, self.hide_annotation_btn]),
                self.threshold_slider,
                self.small_object_size,
            ],
            layout=frame_layout,
        )

        preprocessing_column = VBox(
            [
                self.grid_widgets,
                self.contrast_widgets,
            ],
            layout=fixed_frame_layout,
        )

        annotation_column = VBox(
            [
                self.manual_annotation_widgets,
                self.auto_annotation_widgets,
            ],
            layout=fixed_frame_layout,
        )

        control_panel = HBox([preprocessing_column, annotation_column])

        super().__init__([control_panel, self.canvas_widget])

    def get_annotation_mask(self) -> np.ndarray:
        """Get the annotation mask.

        The mask is a numpy array with the same shape as the original image,
        where the pixels with value 1 correspond to the bacterial colony, and 0 otherwise.
        """
        return self.mask

    def get_metrics(
        self,
        brightness_mode: Literal[
            "luminance", "luminance-inverse", "red", "green", "blue"
        ] = "luminance-inverse",
    ):
        """Get the metrics of the image after the annotation.

        Args:
            brightness_mode: mode to calculate the brightness,
                one of 'luminance', 'luminance-inverse', 'red', 'green', 'blue'.
                Default is 'luminance-inverse' used in the paper.
        """
        return get_summary_metrics(
            self.cut_img(self.original_img),
            self.mask,
            grid_x=self.canvas_widget.grid_num_x,
            grid_y=self.canvas_widget.grid_num_y,
            mode=brightness_mode,
        )

    def cut_img(self, img: np.ndarray) -> np.ndarray:
        """Cut the image according to the current padding settings.

        Args:
            img: The input image to be cut.

        Returns:
            The cut image.
        """
        w = self.canvas_widget
        top, bottom, left, right = map(
            lambda x: int(round(x)), (w.pad_top, w.pad_bottom, w.pad_left, w.pad_right)
        )
        img = img[bottom:-top, left:-right]
        return img

    def apply_auto_annotation(self, *args):
        """Apply automatic annotation based on current threshold and size settings."""
        # self.annotate_btn.description = str(btn.value)
        t = self.threshold_slider.value
        s = self.small_object_size.value
        self.disable_widgets()

        processed_img, mask = segment_by_thresholding(self._preprocessed_img, t, s)

        self._saved_mask = mask
        self.canvas_widget.set_annotation(mask)
        self.disable_widgets(False)

    def _set_image(self, img=None):
        if img is None:
            img = self._preprocessed_img
        self.canvas_widget.set_image(img, cmap=self.cmap_down.value)

    def _update_image_view(self, *args):
        self._set_image()

    def disable_widgets(self, value: bool = True):
        """Enable or disable all control widgets.

        Args:
            value: True to disable widgets, False to enable them.
        """
        for widget in self.control_widgets + [self.hide_annotation_btn, self.clahe_btn]:
            widget.disabled = value

    def _hide_annotation_clicked(self, *args):
        hide = self.hide_annotation_btn.value

        if hide:
            self.hide_annotation_btn.description = "Show annotation"
            self._prev_draw_btn_state = self.draw_mode_btns.value
            self.draw_mode_btns.value = "Off"

            self._saved_mask = self.mask

            for widget in self.control_widgets:
                widget.disabled = True

            self.canvas_widget.set_annotation(np.zeros_like(self._saved_mask))
        else:
            self.hide_annotation_btn.description = "Hide annotation"

            for widget in self.control_widgets:
                widget.disabled = False
            self.draw_mode_btns.value = self._prev_draw_btn_state

            self.canvas_widget.set_annotation(self._saved_mask)

    def _update_preprocessed_image(self, *args):
        use_clahe = self.clahe_btn.value
        clahe_limit = self.clahe_limit_slider.value
        subtract_background = self.subtract_background_btn.value

        self._preprocessed_img = working_preprocessing(
            self.cut_img(self.original_img),
            use_clahe=use_clahe,
            clahe_limit=clahe_limit,
            subtract_background=subtract_background,
        )

        self._set_image(self._preprocessed_img)

    def _change_grid_btn_clicked(self, *args):
        change_grid_mode = self.change_grid_btn.value
        self.canvas_widget.change_grid = change_grid_mode

        if change_grid_mode:
            self._set_image(self.original_img)
        else:
            self._update_preprocessed_image()

    @property
    def mask(self):
        """Get the current annotation mask as a binary image.

        Returns:
            A binary mask where 1 indicates annotated pixels.
        """
        mask = self.canvas_widget.get_annotation()
        mask = mask.sum(-1)
        mask[mask > 0] = 1.0
        return mask

    def init_connections(self):
        """Initialize all widget connections and event observers."""
        # connect widgets
        self.clahe_btn.observe(self._update_preprocessed_image, "value")
        self.clahe_limit_slider.observe(self._update_preprocessed_image, "value")
        self.subtract_background_btn.observe(self._update_preprocessed_image, "value")
        self.cmap_down.observe(self._update_image_view, "value")

        self.annotate_btn.on_click(self.apply_auto_annotation)
        self.hide_annotation_btn.observe(self._hide_annotation_clicked, "value")
        self.threshold_slider.observe(self.apply_auto_annotation, "value")
        self.small_object_size.observe(self.apply_auto_annotation, "value")
        self.change_grid_btn.observe(self._change_grid_btn_clicked, "value")

        # link properties
        ipywidgets.link((self.draw_mode_btns, "value"), (self.canvas_widget, "mode"))
        ipywidgets.link((self.brush_size_slider, "value"), (self.canvas_widget, "brush_size"))
        ipywidgets.link((self.x_grid_size_slider, "value"), (self.canvas_widget, "grid_num_x"))
        ipywidgets.link((self.y_grid_size_slider, "value"), (self.canvas_widget, "grid_num_y"))
        ipywidgets.link((self.top_pad_slider, "value"), (self.canvas_widget, "pad_top"))
        ipywidgets.link((self.bottom_pad_slider, "value"), (self.canvas_widget, "pad_bottom"))

        ipywidgets.link((self.right_pad_slider, "value"), (self.canvas_widget, "pad_right"))
        ipywidgets.link((self.left_pad_slider, "value"), (self.canvas_widget, "pad_left"))
        ipywidgets.link((self.hide_grid_btn, "value"), (self.canvas_widget, "hide_grid"))



"""Tests for the bactovision.image_processing module."""

import numpy as np
import pytest

from bactovision.image_processing import (
    add_convex_hulls,
    clahe,
    get_summary_metrics,
    normalize_image,
    preprocess_image,
    segment_by_thresholding,
)

from .utils import (
    create_test_binary_mask,
    create_test_color_image,
    create_test_grayscale_image,
    create_test_labeled_mask,
)


def test_normalize_image():
    """Test the normalize_image function."""
    # Create a simple test array
    test_array = np.array([[0, 50, 100], [150, 200, 255]], dtype=np.uint8)

    # Normalize the array
    normalized = normalize_image(test_array)

    # Check that the values are between 0 and 1
    assert np.min(normalized) == 0
    assert np.max(normalized) == 1

    # Check that the original values are proportionally maintained
    assert normalized[0, 0] == 0
    assert normalized[1, 2] == 1
    assert normalized[0, 1] == pytest.approx(50 / 255)


def test_normalize_image_constant():
    """Test normalize_image with a constant image."""
    # Create a constant array
    test_array = np.ones((5, 5)) * 100

    # Normalize the array
    normalized = normalize_image(test_array)

    # Check that the result is all zeros (as per the function definition)
    assert np.all(normalized == 0)


def test_segment_by_thresholding():
    """Test the segment_by_thresholding function."""
    # Create a test image
    img = create_test_grayscale_image()

    # Apply thresholding
    _, mask = segment_by_thresholding(img, t=0.8, s=0.5)

    # Check that the mask has some non-zero values
    assert np.max(mask) > 0

    # Check that the mask has the same shape as the input
    assert mask.shape == img.shape


def test_add_convex_hulls():
    """Test the add_convex_hulls function."""
    # Create a simple labeled mask
    mask = create_test_labeled_mask()

    # Apply convex hull
    result = add_convex_hulls(mask)

    # Check the result has the same shape
    assert result.shape == mask.shape

    # Check that the result has values 1 and 2
    assert set(np.unique(result)) == {0, 1, 2}

    # Check that the convex hull filled in some additional pixels
    assert np.sum(result > 0) >= np.sum(mask > 0)


def test_clahe():
    """Test the CLAHE function."""
    # Create a simple test image
    img = np.array([[0, 50], [150, 255]], dtype=np.uint16)

    # Apply CLAHE
    result = clahe(img, limit=100)

    # Check the result has the same shape
    assert result.shape == img.shape

    # Check that the result is a numpy array
    assert isinstance(result, np.ndarray)


def test_preprocess_image():
    """Test the preprocess_image function."""
    # Create a test image
    img = create_test_color_image()

    # Preprocess the image
    processed = preprocess_image(img)

    # Check that the result is a 2D array
    assert processed.ndim == 2

    # Check that the values are between 0 and 1
    assert np.min(processed) >= 0
    assert np.max(processed) <= 1

    # Test with CLAHE
    processed_clahe = preprocess_image(img, use_clahe=True)
    assert processed_clahe.ndim == 2
    assert np.min(processed_clahe) >= 0
    assert np.max(processed_clahe) <= 1


def test_get_summary_metrics():
    """Test the get_summary_metrics function."""
    # Create a test image and mask
    img = create_test_color_image()
    mask = create_test_binary_mask(size=img.shape[:2])

    # Get metrics
    metrics = get_summary_metrics(img, mask, grid_x=2, grid_y=2, mode="luminance")

    # Check that the metrics contain the expected keys
    assert "intergal_opacity" in metrics
    assert "average_opacity" in metrics
    assert "relative_area" in metrics
    assert "num_pixels" in metrics

    # Check the shape of the metrics
    assert metrics["relative_area"].shape == (2, 2)



"""Setup script for BactoVision package."""

import pathlib

from setuptools import find_packages, setup

# Get the long description from the README file
root_dir = pathlib.Path(__file__).parent.resolve()
long_description = (root_dir / "README.md").read_text(encoding="utf-8")

setup(
    name="bactovision",
    version="0.1.1",
    description="A package for bacterial image processing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Vladimir Starostin",
    author_email="vladimir.starostin@uni-tuebingen.de",
    url="https://github.com/StarostinV/bactovision",
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    install_requires=[
        "numpy>=1.20.0,<2.0.0",
        "scipy>=1.8.0,<1.15.0",
        "scikit-image>=0.18.0",
        "Pillow>=9.0.0",
        "opencv-python>=4.5.0",
        "anywidget>=0.1.0",
        "traitlets>=5.0.0",
        "matplotlib>=3.5.0",
        "jupyterlab>=3.0.0",
    ],
    python_requires=">=3.8",
)



"""CanvasWidget is a jupyter widget that allows you to draw on an image."""

import base64
from io import BytesIO
from pathlib import Path

import anywidget
import matplotlib.cm as cm
import numpy as np
import traitlets
from PIL import Image

from bactovision.image_processing import normalize_image


class CanvasWidget(anywidget.AnyWidget):
    """CanvasWidget is a jupyter widget that allows you to draw on an image."""

    # js code
    _esm = Path(__file__).parent / "canvas.js"

    # image
    image_data = traitlets.Unicode().tag(sync=True)

    # annotation
    brush_size = traitlets.Float(5).tag(sync=True)  # Default brush size set to 10
    mode = traitlets.Unicode("add-segments").tag(sync=True)
    annotated_image = traitlets.Unicode("").tag(sync=True)

    # grid
    grid_num_x = traitlets.Int(12).tag(sync=True)
    grid_num_y = traitlets.Int(8).tag(sync=True)
    hide_grid = traitlets.Bool().tag(sync=True)
    change_grid = traitlets.Bool().tag(sync=True)

    # grid padding. These parameters should not be controlled by setting values,
    # since the change will not be reflected on a canvas immediately.
    pad_top = traitlets.Float(50).tag(sync=True)
    pad_bottom = traitlets.Float(60).tag(sync=True)
    pad_left = traitlets.Float(90).tag(sync=True)
    pad_right = traitlets.Float(50).tag(sync=True)

    def set_image(self, image_array: np.array, cmap: str = "YlGnBu") -> None:
        """Set the image to be displayed in the widget.

        Args:
            image_array: numpy array of the image.
            cmap: colormap to use for the image.
        """
        self.image_data = array2str(image_array, cmap=cmap)

    def set_annotation(self, mask: np.array) -> None:
        """Set the annotation mask to be displayed in the widget.

        Args:
            mask: numpy array of the mask.
        """
        self.annotated_image = array2str(mask2red(mask))

    def get_annotation(self) -> np.array:
        """Get the annotation mask from the widget.

        Returns:
            numpy array of the mask.
        """
        return str2array(self.annotated_image)

    def get_grid_dict(self) -> dict:
        """Get the grid parameters from the widget.

        Returns:
            dictionary of the grid parameters.
        """
        return {
            "num_x": int(self.grid_num_x),
            "num_y": int(self.grid_num_y),
            "pad_top": float(self.pad_top),
            "pad_bottom": float(self.pad_bottom),
            "pad_left": float(self.pad_left),
            "pad_right": float(self.pad_right),
        }


def str2array(base64_str: str) -> np.ndarray:
    """Convert a base64 string to a numpy array.

    Args:
        base64_str: base64 string of the image.

    Returns:
        A numpy array representing the decoded image.
    """
    base64_str = base64_str.split(",")[1] if "," in base64_str else base64_str
    image_bytes = base64.b64decode(base64_str)
    return np.array(Image.open(BytesIO(image_bytes)))


def array2str(arr: np.ndarray, cmap: str = "YlGnBu") -> str:
    """Convert a numpy array to a base64 string.

    Args:
        arr: numpy array of the image.
        cmap: colormap to use for the image.

    Returns:
        A base64 encoded string of the image.
    """
    if len(arr.shape) == 2:
        arr = cm.get_cmap(cmap)(arr)

    arr = (normalize_image(arr) * 255).astype("uint8")

    assert len(arr.shape) == 3

    mode = "RGB" if arr.shape[-1] == 3 else "RGBA"

    image = Image.fromarray(arr, mode)  # Use 'RGBA' if your array includes an alpha channel

    # Save the PIL Image to a bytes buffer
    buffer = BytesIO()
    image.save(
        buffer, format="PNG"
    )  # You can change 'PNG' to 'JPEG' or other formats depending on your needs
    buffer.seek(0)

    # Encode the bytes buffer to base64
    image_base64 = base64.b64encode(buffer.read())

    # Decode bytes to a string and format for use as a data URL if needed
    image_base64_str = "data:image/png;base64," + image_base64.decode()

    return image_base64_str


def mask2red(mask: np.ndarray) -> np.ndarray:
    """Convert a mask to a red image.

    Args:
        mask: numpy array of the mask.

    Returns:
        numpy array of the red image.
    """
    annotation = ((mask > 0).astype(np.uint8) * 255)[..., None].repeat(4, 2)
    annotation[(mask > 0), 0] = 255
    annotation[(mask > 0), 3] = 255 // 2

    annotation[(mask == 0), 0] = 0
    annotation[(mask == 0), 3] = 0

    annotation[..., [1, 2]] = 0
    return annotation
